# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_annotate.ipynb (unless otherwise specified).

__all__ = ['create_label_studio_json', 'process_labels', 'load_annotations_csv', 'check_download_df_match', 'load_df',
           'load_completions', 'anno_sample_merge', 'nnAnnotations']

# Cell
from .core import *

# Cell
from tqdm.notebook import trange, tqdm
from toolz.itertoolz import count
import pandas as pd
from pandas import json_normalize
import simplejson as json
import requests
import re
from datetime import datetime
from glob import glob
from pathlib import Path

# Cell
import nnanno
from typing import Union, Optional, Type

# Cell
from .sample import *

# Cell
def create_label_studio_json(
    sample: Union[pd.DataFrame, Type[nnSampler]],
    fname: Union[str, Path, None] = None,
    original: bool = True,
    pct: Optional[int] = None,
    size: Optional[tuple] = None,
    preserve_asp_ratio: bool = True,
):
    """create a json file which can be used to upload tasks to label studio"""
    if fname and Path(fname).exists():
        raise FileExistsError(f"{fname} already exists")
    if fname is None:
        today = datetime.today()
        time_stamp = today.strftime("%Y_%d_%m_%H_%M")
        fname = f"{time_stamp}_tasks.json"
    if type(sample) == nnanno.sample.nnSampler:
        try:
            sample = sample.sample.copy()
        except AttributeError as e:
            print(f"{sample} doesn't have a sample associated with it")
    else:
        sample = sample.copy()
    sample["image"] = sample.apply(
        lambda x: iiif_df_apply(
            x,
            original=original,
            pct=pct,
            size=size,
            preserve_asp_ratio=preserve_asp_ratio,
        ),
        axis=1,
    )
    label_studio_json = sample.apply(lambda x: x.to_dict(), axis=1).to_list()
    with open(fname, "w") as f:
        json.dump(label_studio_json, f, ignore_nan=True)

# Cell
def process_labels(x):
    try:
        x = "|".join(eval(x)["choices"])
    except:
        NameError
    return x

# Cell
def load_annotations_csv(csv: Union[str, Path], kind="classification"):
    if kind == "classification":
        df = pd.read_csv(csv, converters={"box": eval})
        df["label"] = df["choice"]
        return df
    if kind == "label":
        df = pd.read_csv(csv, converters={"box": eval})
        df["label"] = df["choice"].apply(process_labels)
        return df

# Cell
def check_download_df_match(dl_folder: Union[Path, str], df: pd.DataFrame) -> str:
    im_count = count(
        f for f in Path(dl_folder).iterdir() if f.suffix in image_extensions
    )
    if type(df) == pd.core.frame.DataFrame:
        if len(df) == im_count:
            print(
                f"Length of DataFrame {len(df)} and number of images in {dl_folder} {im_count} match",
                "\U0001F600",
            )
        if len(df) != im_count:
            print(
                f"Length of DataFrame {len(df)} and number of images in {dl_folder} {im_count} do not match",
                "\U0001F615",
            )

# Cell
def load_df(json_file: Union[str, Path]):
    with open(json_file) as f:
        data = json.load(f)
        df = json_normalize(data, record_path=["completions"], meta=["data"])
        # df['result'] = df['result'].apply(lambda x: return_choice(x[0]) if len([x][0]) ==1 else x)
        df["result"] = df["result"].apply(
            lambda x: x[0]["value"]["choices"] if len([x][0]) == 1 else x
        )
        return df

# Cell
def load_completions(path: Union[str, Path]):
    filenames = glob(f"{path}/completions/*.json")
    dataframes = [load_df(f) for f in filenames]
    return pd.concat(dataframes)

# Internal Cell
def _df_to_csv(df, out_fn):
    df[["data", "result"]].to_csv(
        out_fn,
        header=[
            "file",
            "label",
        ],
        index=False,
    )

# Internal Cell
def _df_to_json(df, out_fn):
    df[["data", "value.choices"]].to_json(out_fn)

# Internal Cell
def _df_to_pkl(df, out_fn):
    df.to_pickle(out_fn)

# Internal Cell
def get_og_filepath(x):
    """
    Transforms a filepaths from processed ImageStudio format back to the Orginal Newspaper Navigator filepath  format
    """
    b, m, e = re.split("(_data_)", x)
    m = m.replace("_", "/")
    e = re.split("(\d{3}_\d{1}_\d{2}.jpg)", e)
    return b + m + e[0].replace("_", "/") + e[1]

# Cell
def anno_sample_merge(
    sample_df: pd.DataFrame, annotation_df: pd.DataFrame
) -> pd.DataFrame:
    """anno_sample_merge merges a DataFrame containing a sample
    from Newspaper Navigator and a DataFrame containing annotations

    Parameters
    ----------
    sample_df : pd.DataFrame
        A Pandas DataFrame which holds a sample from Newspaper Navigator Generated by `sample.nnSample()`
    annotation_df : pd.DataFrame
        A pandas DataFrame containing annotations loaded via the `annotate.nnAnnotations` class


    Returns
    -------
    pd.DataFrame
        A new DataFrame which merges the two input DataFrames
    """
    sample_df, annotation_df = sample_df.copy(), annotation_df.copy()
    annotation_df["id"] = annotation_df["data"].map(lambda x: get_og_filepath(x))
    return sample_df.merge(annotation_df, left_on="filepath", right_on="id")

# Cell


class nnAnnotations:
    def __init__(self, df):
        self.annotation_df = df
        self.labels = df["result"].unique()
        self.label_counts = df["result"].value_counts()

    def __repr__(self):
        return f"{self.__class__.__name__}" f" #annotations:{len(self.annotation_df)}"

    @classmethod
    def from_completions(cls, path, kind, drop_dupes=True, sample_df=None):
        df = load_completions(path)
        df = df.reset_index(drop=True)  # add index
        df["data"] = df["data"].map(lambda x: x["image"])
        df["data"] = df["data"].map(lambda x: x.split("?")[0])
        df["data"] = df["data"].apply(lambda x: Path(x).name)
        if any(
            df["data"].str.contains("-")
        ):  # removes labelstudio hash from data loaded via web interface
            df["data"] = df["data"].str.split("-", expand=True)[1]
        if drop_dupes:
            df = df.drop_duplicates(subset="data", keep="last")
        if kind == "classification":
            empty_rows = df[df["result"].apply(lambda x: len(x) == 0)].index
            df = df.drop(empty_rows)
            df["result"] = df["result"].map(lambda x: x[0])
        if kind == "label":
            df["result"] = df["result"].map(
                lambda x: "|".join(map(str, x)) if len(x) >= 1 else x
            )
            df["result"] = df["result"].map(lambda x: "" if len(x) == 0 else x)
        return cls(df)

    def merge_sample(self, sample_df):
        self.merged_df = anno_sample_merge(sample_df, self.annotation_df)

    def export_merged(self, out_fn):
        self.merged_df.to_csv(out_fn)

    def export_annotations(self, out_fn):
        df = self.annotation_df
        if not Path(out_fn).exists():
            Path(out_fn).touch()
        suffix = Path(out_fn).suffix
        if suffix == ".csv":
            _df_to_csv(df, out_fn)
        if suffix == ".json":
            _df_to_json(df, out_fn)
        if suffix == ".pkl":
            _df_to_pkl(df, out_fn)


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"create_session": "00_core.ipynb",
         "create_cached_session": "00_core.ipynb",
         "get_max_workers": "00_core.ipynb",
         "load_url_image": "00_core.ipynb",
         "save_image": "00_core.ipynb",
         "download_image": "00_core.ipynb",
         "parse_box": "00_core.ipynb",
         "create_iiif_url": "00_core.ipynb",
         "iiif_df_apply": "00_core.ipynb",
         "image_extensions": "00_core.ipynb",
         "bytesto": "00_core.ipynb",
         "get_json_url": "01_sample.ipynb",
         "load_json": "01_sample.ipynb",
         "count_json_iter": "01_sample.ipynb",
         "get_year_size": "01_sample.ipynb",
         "get_year_sizes": "01_sample.ipynb",
         "get_all_year_sizes": "01_sample.ipynb",
         "sample_stream": "01_sample.ipynb",
         "calc_frac_size": "01_sample.ipynb",
         "calc_year_from_total": "01_sample.ipynb",
         "reduce_df_memory": "01_sample.ipynb",
         "nnSampler": "01_sample.ipynb",
         "sample_year": "01_sample.ipynb",
         "create_sample": "01_sample.ipynb",
         "download_sample": "01_sample.ipynb",
         "create_label_studio_json": "02_annotate.ipynb",
         "process_labels": "02_annotate.ipynb",
         "load_annotations_csv": "02_annotate.ipynb",
         "check_download_df_match": "02_annotate.ipynb",
         "load_df": "02_annotate.ipynb",
         "load_completions": "02_annotate.ipynb",
         "get_og_filepath": "02_annotate.ipynb",
         "anno_sample_merge": "02_annotate.ipynb",
         "nnAnnotations": "02_annotate.ipynb",
         "nnPredict": "03_inference.ipynb",
         "predict_from_sample_df": "03_inference.ipynb",
         "predict_sample": "03_inference.ipynb",
         "predict": "03_inference.ipynb"}

modules = ["core.py",
           "sample.py",
           "annotate.py",
           "inference.py"]

doc_url = "https://Living-with-machines.github.io/nnanno/"

git_url = "https://github.com/Living-with-machines/nnanno/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_core.ipynb (unless otherwise specified).

__all__ = ['create_session', 'create_cached_session', 'get_max_workers', 'load_url_image', 'save_image',
           'download_image', 'parse_box', 'create_iiif_url', 'iiif_df_apply', 'image_extensions', 'bytesto']

# Cell
import math
import mimetypes
import pathlib
from pathlib import Path
import PIL
from PIL import Image  # UnidentifiedImageError
import io
import pandas as pd
import multiprocessing
import requests
import requests_cache
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from typing import (
    Any,
    Optional,
    Union,
    Dict,
    List,
    Tuple,
    Set,
    Iterable,
)

# Cell
def create_session() -> requests.sessions.Session:
    """returns a requests session"""
    retry_strategy = Retry(total=80)
    adapter = HTTPAdapter(max_retries=retry_strategy)
    session = requests.Session()
    session.mount("https://", adapter)
    session.mount("http://", adapter)
    return session

# Cell
def create_cached_session() -> requests_cache.CachedSession:
    """Creates a session which caches requests"""
    retry_strategy = Retry(total=80)
    adapter = HTTPAdapter(max_retries=retry_strategy)
    session = requests_cache.CachedSession("url_cache")
    session.mount("http://", adapter)
    return session

# Cell
def get_max_workers(data=None) -> int:
    """Returns int to pass to max_workers based on len of `data` if available or `cpu_count()`"""
    if data is not None and hasattr(data, "__len__"):
        return min(multiprocessing.cpu_count(), len(data))
    else:
        return multiprocessing.cpu_count()

# Cell
def load_url_image(url: str, mode="RGB") -> Union[PIL.Image.Image, None]:
    "Attempts to load an image from `url` returns `None` if request times out or no image at `url`"
    im = None
    session = create_session()
    with session.get(url, timeout=(30)) as r:
        if r:
            try:
                im = (Image.open(io.BytesIO(r.content))).convert(mode)
            except:
                pass
        return im

# Cell
def save_image(
    im: PIL.Image.Image, fname: str, out_dir: Union[str, pathlib.Path] = "."
):
    """Saves `im` as `fname` to `out_dir`"""
    out_path = Path(f"{out_dir}/{fname}")
    im.save(out_path)

# Cell
def download_image(
    url: str, fname: str, out_dir: Union[str, pathlib.Path] = "."
) -> None:
    """
    Attempts to load image from `url` and save as `fname` to `out_dir`
    Returns `None` if bad URL or request timesout
    """
    im = load_url_image(url)
    if im:
        save_image(im, fname, out_dir)
    else:
        return None

# Cell
def parse_box(box: Union[Tuple, List]) -> Tuple[float, float, float, float]:
    """Parses the `box` value from Newspaper Navigator data to prepre for IIIF request"""
    box_x1, box_x2, box_y1, box_y2 = box
    x = math.floor(box_x1 * 10000) / 100.0
    y = math.ceil(box_x2 * 10000) / 100.0
    w = math.ceil((box_y1 - box_x1) * 10000) / 100.0
    h = math.ceil((box_y2 - box_x2) * 10000) / 100.0
    return x, y, w, h

# Cell
def create_iiif_url(
    box: Union[Tuple, List],
    url: str,
    original: bool = False,
    pct: int = None,
    size: tuple = None,
    preserve_asp_ratio: bool = True,
) -> Union[str, None]:
    """Returns a IIIF URL from bounding box and URL"""

    x, y, w, h = parse_box(box)
    url_coordinates = "pct:" + str(x) + "," + str(y) + "," + str(w) + "," + str(h)
    url_chronam_path = "%2F".join(url.split("/")[4:10]) + ".jp2"

    url_prefix = "https://chroniclingamerica.loc.gov/iiif/2"
    if original and not size and not pct:
        url_suffix_full = "pct:100/0/default.jpg"
        return "/".join(
            [url_prefix, url_chronam_path, url_coordinates, url_suffix_full]
        )
    if pct:
        pct_downsampled = f"pct:{pct}/0/default.jpg"
        return "/".join(
            [url_prefix, url_chronam_path, url_coordinates, pct_downsampled]
        )
    if size and preserve_asp_ratio:
        return "/".join(
            [
                url_prefix,
                url_chronam_path,
                url_coordinates,
                f"!{size[0]},{size[1]}/0/default.jpg",
            ]
        )
    if size:
        return "/".join(
            [
                url_prefix,
                url_chronam_path,
                url_coordinates,
                f"{size[0]},{size[1]}/0/default.jpg",
            ]
        )

# Cell
def iiif_df_apply(
    row,
    original: bool = False,
    pct: int = 50,
    size: tuple = None,
    preserve_asp_ratio: bool = True,
):
    """Creates IIIF urls from a pandas DataFrame containing newspaper navigator data"""
    return create_iiif_url(
        row["box"],
        row["url"],
        original=original,
        pct=pct,
        size=size,
        preserve_asp_ratio=preserve_asp_ratio,
    )

# Cell
image_extensions = set(
    k for k, v in mimetypes.types_map.items() if v.startswith("image/")
)

# Cell
def bytesto(bytes, to: str, bsize: int = 1024) -> float:
    """Takes bytes and returns value convereted to `to`"""
    a = {"k": 1, "m": 2, "g": 3, "t": 4, "p": 5, "e": 6}
    r = float(bytes)
    return bytes / (bsize ** a[to])


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_sample.ipynb (unless otherwise specified).

__all__ = ['get_json_url', 'load_json', 'count_json_iter', 'get_year_size', 'get_year_sizes', 'get_all_year_sizes',
           'sample_stream', 'calc_frac_size', 'calc_year_from_total', 'reduce_df_memory', 'nnSampler', 'sample_year',
           'create_sample', 'download_sample']

# Cell
from .core import *

# Cell
# TODO tidy imports
# sys
import io
import shutil
import pkg_resources
from pathlib import Path
from datetime import datetime

# other
from tqdm.auto import trange, tqdm
import requests
import ijson
import functools
import math
from cytoolz import dicttoolz, itertoolz
import random
import json
from PIL import Image

import concurrent.futures
import numpy as np
import itertools
from pandas import json_normalize
import pandas as pd
from functools import partial
import numpy as np
from fastcore.foundation import patch_to

# Cell

import PIL
from typing import (
    Any,
    Optional,
    Union,
    Dict,
    List,
    Tuple,
    Set,
    Iterable,
)

# Cell
def get_json_url(year: Union[str, int], kind: str = "photos") -> str:
    """Returns url for the json data from news-navigator for given `year` and `kind`"""
    return f"https://news-navigator.labs.loc.gov/prepackaged/{year}_{kind}.json"

# Cell
def load_json(url: str) -> Dict[str, Any]:
    """Returns json loaded from `url`"""
    with requests.get(url, timeout=2) as r:
        r.raise_for_status()
        return json.loads(r.content)

# Cell
@functools.lru_cache(256)
def count_json_iter(url: str, session=None) -> int:
    """Returns count of objects in url json file using an iterator to avoid loading json into memory"""
    if not session:
        session = create_cached_session()
    with session.get(url, timeout=60) as r:
        r.raise_for_status()
        if r:
            objects = ijson.items(r.content, "item")
            count = itertoolz.count(iter(objects))
        else:
            count = np.nan
    return count

# Cell
@functools.lru_cache(256)
def get_year_size(year: Union[int, str], kind: str) -> dict:
    """returns size of a json dataset for a given year and kind
    results are cached
    Parameters
    ----------
    year : Union[int,str]
        year from newspaper navigator
    kind : str
        {'ads', 'photos', 'maps', 'illustrations', 'comics', 'cartoons', 'headlines'}
    Returns
    -------
    size :dict
        returns a dict with year as a key and size as value
    """
    session = None
    dset_size = {}
    url = get_json_url(year, kind)
    if kind == "ads" and int(year) >= 1870 or (kind == "headlines"):
        session = create_session()
    dset_size[str(year)] = count_json_iter(url, session)
    return dset_size

# Cell
@functools.lru_cache(512)
def get_year_sizes(kind: str, start: int = 1850, end: int = 1950, step: int = 5):
    """
    Returns the sizes for json data files for `kind` between year `start` and `end`
    with step size 'step'

    Parameters:
    kind (str): kind of image from news-navigator:
    {'ads', 'photos', 'maps', 'illustrations', 'comics', 'cartoons', 'headlines'}

    Returns:
    Pandas.DataFrame: holding data from input json url
    """
    futures = []
    years = range(start, end + 1, step)
    max_workers = get_max_workers(years)
    with tqdm(total=len(years)) as progress:
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            for year in years:
                future = executor.submit(get_year_size, year, kind)
                future.add_done_callback(lambda p: progress.update())
                futures.append(future)
        results = [future.result() for future in futures]
        dset_size = {k: v for d in results for k, v in d.items()}
    return pd.DataFrame.from_dict(dset_size, orient="index", columns=[f"{kind}_count"])

# Cell
def get_all_year_sizes(
    start: int = 1850, end: int = 1950, step: int = 1, save: bool = True
):
    """
    Returns a dataframe with number of counts from year `start` to `end`
    """
    kinds = [
        "ads",
        "photos",
        "maps",
        "illustrations",
        "comics",
        "cartoons",
        "headlines",
    ]
    dfs = []
    for kind in tqdm(kinds):
        df = get_year_sizes(kind, start=start, end=end, step=step)
        dfs.append(df)
    df = pd.concat(dfs, axis=1)
    df["total"] = df.sum(axis=1)
    if save:
        df.to_csv("all_year_sizes.csv")
    return df

# Cell
def sample_stream(stream, k: int):
    """
    Return a random sample of k elements drawn without replacement from stream.
    Designed to be used when the elements of stream cannot easily fit into memory.
    """
    r = np.array(list(itertools.islice(stream, k)))
    for t, x in enumerate(stream, k + 1):
        i = np.random.randint(1, t + 1)
        if i <= k:
            r[i - 1] = x
    return r

# Cell
@functools.lru_cache(1024)
def calc_frac_size(url, frac, session=None):
    "returns fraction size from a json stream"
    return round(count_json_iter(url, session) * frac)

# Cell
def calc_year_from_total(total, start, end, step):
    "Calculate size of a year sample based on a total sample size"
    return max(1, round(total / (((end - start) + 1) / step)))

# Cell
def reduce_df_memory(df):
    return df.astype(
        {
            "score": "float64",
            "page_seq_num": "int32",
            "batch": "category",
            "box": "object",
            "lccn": "category",
            "page_url": "category",
            "name": "category",
            "publisher": "category",
            "place_of_publication": "category",
            "edition_seq_num": "category",
        }
    )

# Cell
class nnSampler:
    """Sampler for creating samples from Newspaper Navigator data"""

    population = pd.read_csv(
        pkg_resources.resource_stream("nnanno", "data/all_year_counts.csv"), index_col=0
    )

    def __repr__(self):
        return f"{self.__class__.__name__}"

# Cell
def sample_year(kind: str, sample_size: Union[int, float], year: int) -> np.array:
    """samples `sample_size` for `year` and `kind`"""
    url = get_json_url(year, kind)
    if kind == "ads" and int(year) >= 1870 or (kind == "headlines"):
        session = create_session()
    else:
        session = create_cached_session()
    if type(sample_size) is float:
        sample_size = max(1, calc_frac_size(url, sample_size, session))
        if kind == "ads" and int(year) >= 1870 or (kind == "headlines"):
            session = create_session()
        else:
            session = create_cached_session()
    with session.get(get_json_url(year, kind)) as r:
        if r:
            try:
                data = ijson.items(r.content, "item")
                sample_data = sample_stream(iter(data), sample_size)
            except requests.exceptions.RequestException as e:
                sample_data = np.nan
        return sample_data

# Cell
@patch_to(nnSampler)
def create_sample(
    self,
    sample_size: Union[int, float],
    kind: str = "photos",
    start_year: int = 1850,
    end_year: int = 1950,
    step: int = 5,
    year_sample=True,
    save: bool = False,
    reduce_memory=True,
):
    """
    Creates a sample of Newspaper Navigator data for a given set of years and a kind

    Parameters:
    sample_size: int, float
        `sample size` can either be a fixed number or a fraction of the total dataset size
    kind (str): kind of image from news-navigator:
    {'ads', 'photos', 'maps', 'illustrations', 'comics', 'cartoons', 'headlines'}


    Returns:
    Pandas.DataFrame: holding data from input json url
    """

    if not year_sample:
        if type(sample_size) != int:
            raise ValueError(
                f"""type{sample_size} is not an int. Fractions are only supported
                            for sampling by year"""
            )
        sample_size = calc_year_from_total(sample_size, start_year, end_year, step)
    futures = []
    years = range(start_year, end_year + 1, step)
    _year_sample = partial(sample_year, kind, sample_size)
    with tqdm(total=len(years)) as progress:
        workers = get_max_workers(years)
        with concurrent.futures.ThreadPoolExecutor(1) as executor:
            for year in years:
                future = executor.submit(_year_sample, year)
                future.add_done_callback(lambda p: progress.update())
                futures.append(future)
    results = [future.result() for future in futures]
    df = pd.DataFrame.from_dict(list(itertoolz.concat(results)))

    if reduce_memory:
        df = reduce_df_memory(df)
    if save:
        df.to_json(f"{kind}_{start_year}_{end_year}_sample.json")
    self.sample = df
    return df

# Cell
@patch_to(nnSampler)
def download_sample(
    self,
    out_dir: str,
    json_name: Optional[str] = None,
    df: Optional[pd.DataFrame] = None,
    original: bool = True,
    pct: Optional[int] = None,
    size: Optional[tuple] = None,
    preserve_asp_ratio: bool = True,
) -> Union[None]:
    """Download images associated with a sample
    The majority of paramters relate to the options available in a IIIF image request
    see `https://iiif.io/api/image/3.0/#4-image-requests` for further information

    Parameters
    ----------
    out_dir
        The save directory for the images
    json_name

    df
        optional DataFrame containing a sample
    original
        if `True` will download orginal size images via IIIF
    pct
        optional value which scales the size of images requested by `pct`
    size
        a tuple representing `width` by `height`, will be passed to IIIF request
    preserve_asp_ratio
        whether to ask the IIIF request to preserve aspect ratio of image or not

    Returns
    -------
    None
    """

    if df is not None:
        self.download_df = df.copy(deep=True)
    else:
        try:
            self.download_df = self.sample.copy(deep=True)
        except AttributeError as E:
            print(
                "You need to create a sample before downloading, or pass in a previously created "
            )
    self.download_df["iiif_url"] = self.download_df.apply(
        lambda x: iiif_df_apply(
            x,
            original=original,
            pct=pct,
            size=size,
            preserve_asp_ratio=preserve_asp_ratio,
        ),
        axis=1,
    )
    self.download_df["download_image_path"] = self.download_df["filepath"].str.replace(
        "/", "_"
    )

    if not Path(out_dir).exists():
        Path(out_dir).mkdir(parents=True)
    _download_image = lambda x: download_image(
        x.iiif_url, x.download_image_path, out_dir
    )
    with tqdm(total=len(self.download_df)) as progress:
        workers = get_max_workers(self.download_df)
        with concurrent.futures.ThreadPoolExecutor(workers) as executor:
            futures = []
            for tuple_row in self.download_df.itertuples():
                future = executor.submit(_download_image, tuple_row)
                future.add_done_callback(lambda p: progress.update())
                futures.append(future)
            del futures
    if json_name is None:
        today = datetime.today()
        time_stamp = today.strftime("%Y_%d_%m_%H_%M")
        json_name = f"{time_stamp}_{len(self.download_df)}_sample"
    self.download_df.to_json(f"{out_dir}/{json_name}.json")


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_inference.ipynb (unless otherwise specified).

__all__ = ['nnPredict', 'predict_from_sample_df', 'predict_sample', 'predict']

# Cell
import ijson
import pkg_resources
import pandas as pd
from cytoolz import itertoolz
from tqdm.auto import tqdm

from fastcore.foundation import patch_to

# Cell
from .core import *
from .sample import *

# Cell
from typing import (
    Any,
    Optional,
    Union,
    Dict,
    List,
    Tuple,
    Set,
    Iterable,
)
from PIL import Image
import PIL

# Cell
try:
    import fastai
    from fastai.vision.all import *
except ModuleNotFoundError as e:
    print(
        "fastai not found, if you want to use the inference module please install fastai"
    )
    raise e

# Cell
def _filter_replace_none_image(results: List[Optional[PIL.Image.Image]]):
    """
    filters a list containing images and `None`, replaces `None` with dummy images
    returns: original images and dummy images + a location index for the dummy images
    """
    fakeim = Image.fromarray(244 * np.ones((250, 250, 3), np.uint8))
    results = L(results)
    none_image_index = results.argwhere(
        lambda x: x is None
    )  # Gets the index for images which are none
    results[none_image_index] = fakeim  # Replaces None with fakeim
    return results.items, none_image_index

# Internal Cell
def _create_pred_header(fname, dls=None):
    columns = [
        "filepath",
        "pub_date",
        "page_seq_num",
        "edition_seq_num",
        "batch",
        "lccn",
        "box",
        "score",
        "ocr",
        "place_of_publication",
        "geographic_coverage",
        "name",
        "publisher",
        "url",
        "page_url",
        "iiif_url",
        "pred_decoded",
    ]
    if dls:
        columns += list(dls.vocab)
    return pd.DataFrame(columns=columns).to_csv(fname, index=None)

# Internal Cell
def _create_year_csv(out_dir, year, kind, dls=None):
    fname = Path(f"{out_dir}/{year}_{kind}.csv")
    _create_pred_header(fname, dls)
    return fname

# Internal Cell
def _create_year_json(out_dir, year, kind, batch):
    return Path(f"{out_dir}/{year}_{kind}_{batch}.json")

# Internal Cell
def _make_directory(directory, force=False):
    if not force:
        if Path(directory).exists() and len(list(os.scandir(directory))) >= 1:
            raise ValueError(f"{directory} already exists and is not empty")
    Path(directory).mkdir(exist_ok=True, parents=True)

# Cell
class nnPredict:
    """`nnPredict` is used in combination with a trained leaner to run inference on Newspaper Navigator"""

    population = pd.read_csv(
        pkg_resources.resource_stream("nnanno", "data/all_year_counts.csv"), index_col=0
    )

    def __init__(self, learner: fastai.learner, try_gpu: bool = True):
        """Creates an ``nnPredict` instance from `learner`, puts on GPU if `try_gpu` is true and CUDA is avilable"""
        self.learner = learner
        self.try_gpu = try_gpu
        self.dls = learner.dls
        self.decode_map = {v: k for v, k in enumerate(self.learner.dls.vocab)}

    def __repr__(self):
        return f"{self.__class__.__name__} \n" f"learner vocab:{self.learner.dls.vocab}"

# Cell
@patch_to(nnPredict, cls_method=True)
def _get_year_population_size(cls, kind: str, year: Union[str, int]):
    """gets year population size"""
    return cls.population[f"{kind}_count"][year]

# Cell
@patch_to(nnPredict, cls_method=True)
def _get_year_sample_size(cls, kind: str, year: Union[str, int], sample_size):
    """get sample size for year"""
    return (cls._get_year_population_size(kind, year) * sample_size).clip(1).round()

# Cell
@patch_to(nnPredict)
def predict_from_sample_df(
    self, sample_df: pd.DataFrame, bs: int = 16, disable_pbar: bool = False
) -> pd.DataFrame:
    """
    Runs inference on `sample_df` using batch size `bs`, `disable_pbar` controls whether to show progress bar
    Returns a Pandas DataFrame containing orginal dataframe and predictions, with labels taken from `learner.dls.vocab`
    """
    self.sample_df = sample_df
    gpu = False
    if self.try_gpu and torch.cuda.is_available():
        gpu = True
    if gpu:
        self.learner.model = self.learner.model.cuda()

    self.sample_df["iiif_url"] = self.sample_df.apply(
        lambda x: iiif_df_apply(x), axis=1
    )
    dfs = []

    splits = max(1, round(len(self.sample_df) / bs))
    for df in tqdm(
        np.array_split(sample_df, splits),
        disable=disable_pbar,
        leave=False,
        desc="Batch progress",
    ):
        futures = []
        for url in df["iiif_url"].to_list():
            with concurrent.futures.ThreadPoolExecutor(4) as e:
                future = e.submit(load_url_image, url)
                futures.append(future)
        results = [future.result() for future in futures]
        image_list, none_index = _filter_replace_none_image(results)
        im_as_arrays = [np.array(image) for image in image_list]
        # if len(none_index) >0:
        #            tqdm.write(f"{none_index} skipped")
        test_data = self.learner.dls.test_dl(im_as_arrays)
        if gpu:
            test_data.to("cuda")
        with self.learner.no_bar():
            pred_tuple = self.learner.get_preds(dl=test_data, with_decoded=True)
        pred_decoded = L(pred_tuple[2], use_list=True)
        pred_tensor = L(pred_tuple[0], use_list=None)
        pred_decoded[none_index] = np.nan
        pred_tensor[none_index] = np.nan
        df["pred_decoded"] = pred_decoded.items
        df["pred_decoded"] = df["pred_decoded"].astype(float)
        # create an empty df column for each class in dls.vocab
        for c in self.dls.vocab:
            df[f"{c}_prob"] = ""
        # append the tensor predictions to the last `c` colomns of the df
        df.iloc[:, -self.dls.c :] = np.hsplit(
            pred_tensor.items.numpy(), self.dls.c
        )  # split into columns

        dfs.append(df)
    df = pd.concat(dfs)
    df["pred_decoded"] = df["pred_decoded"].map(self.decode_map)
    return df

# Cell
@patch_to(nnPredict)
def predict_sample(
    self,
    kind: str,
    out_dir: str,
    sample_size: Union[int, float],
    bs: int = 16,
    start_year: int = 1850,
    end_year: int = 1950,
    step: int = 1,
    year_sample: bool = True,
    size=None,
    force_dir=False,
):
    """runs inference for a sample `sample_size` of `kind` from newspaper navigator using batch size `bs`"""
    _make_directory(out_dir, force_dir)
    years = range(start_year, end_year + 1, step)
    if type(sample_size) == float:
        total = int(self._get_year_sample_size(kind, list(years), sample_size).sum())
    if type(sample_size) == int:
        total = sample_size * len(years)
    dfs = []
    with tqdm(total=total) as pbar:
        for year in years:
            pbar.set_description(f"Predicting: {year}, total progress")
            sample = sample_year(kind, sample_size, year)
            sample_df = pd.DataFrame.from_records(sample)
            disable_pbar = False
            if len(sample_df) <= 2:
                disable_pbar = True
            pred_df = self.predict_from_sample_df(
                sample_df, bs, disable_pbar=disable_pbar
            )
            pred_df.to_json(
                f"{out_dir}/{year}.json"
            )  # TODO make sure this file is created before attempting to save
            pbar.update(len(pred_df))

# Cell
@patch_to(nnPredict)
def predict(
    self,
    kind: str,
    out_dir: str,
    bs: int = 32,
    start_year: int = 1850,
    end_year: int = 1950,
    step: int = 1,
    size=None,
):
    """predict on the full dataset for a given `kind` from `start_year` until `end_year` using `step` size"""
    _make_directory(out_dir)
    gpu = False
    if self.try_gpu and torch.cuda.is_available():
        gpu = True
    if gpu:
        self.learner.model = self.learner.model.cuda()
    years = range(start_year, end_year + 1, step)
    total = self._get_year_population_size(kind, years).sum()
    with tqdm(total=total) as pbar:
        for year in years:
            out_fn = _create_year_csv(out_dir, year, kind, self.learner.dls)
            pbar.set_description(f"Predicting: {year}, total progress")
            if kind == ("ads" and int(year) >= 1870) or (kind == "headlines"):
                s = create_session()
            else:
                s = create_cached_session()
            with s.get(get_json_url(year, kind), timeout=60) as r:
                data = ijson.items(r.content, "item")
                batches = itertoolz.partition_all(bs, iter(data))
                year_total = self._get_year_population_size(kind, year)
                if (year_total // bs) <= 1:
                    disable_p_bar = True
                for i, batch in enumerate(
                    tqdm(
                        batches,
                        total=round(year_total // bs),
                        leave=False,
                        desc="Batch Progress",
                    )
                ):
                    df = pd.DataFrame(batch)
                    df["iiif_url"] = df.apply(lambda x: iiif_df_apply(x), axis=1)
                    futures = []
                    workers = get_max_workers(df)
                    for iif_url in df["iiif_url"].values:
                        with concurrent.futures.ThreadPoolExecutor(workers) as e:
                            future = e.submit(load_url_image, iif_url)
                            futures.append(future)
                    results = [future.result() for future in futures]
                    image_list, none_index = _filter_replace_none_image(results)
                    im_as_arrays = [np.array(image) for image in image_list]
                    if len(none_index) > 0:
                        tqdm.write(f"{none_index} skipped")
                    else:
                        pass
                    test_data = self.learner.dls.test_dl(im_as_arrays)
                    with self.learner.no_bar():
                        pred_tuple = self.learner.get_preds(
                            dl=test_data, with_decoded=True
                        )
                    pred_decoded = L(pred_tuple[2], use_list=True)
                    pred_tensor = L(pred_tuple[0], use_list=None)
                    pred_decoded[none_index] = np.nan
                    pred_tensor[none_index] = np.nan
                    df["pred_decoded"] = pred_decoded.items
                    df["pred_decoded"] = df["pred_decoded"].astype(float)
                    # df['pred_decoded'].map({v: k for v, k in enumerate(self.learner.dls.vocab)})
                    # create an empty df column for each class in dls.vocab
                    for c in self.dls.vocab:
                        df[f"{c}_prob"] = ""
                    # append the tensor predictions to the last `c` columns of the df
                    df.iloc[:, -self.dls.c :] = np.hsplit(
                        pred_tensor.numpy(), self.dls.c
                    )  # split into columns
                    df.to_csv(out_fn, header=None, index=None, mode="a")
                    pbar.update(bs)